import kiwigrad
print(kiwigrad.__version__)
from kiwigrad import *
from sklearn.datasets import make_moons
import numpy as np

X, y = make_moons(n_samples=100, noise=0.1)
print(X.shape, type(X))
print(y.shape, type(y))

class PotNet(MLP):
    def __init__(self):
        self.layers = [
            Layer(nin=2, nout=16, bias=True, activation="relu"),
            Layer(nin=16, nout=16, bias=True, activation="relu"),
            Layer(nin=16, nout=1, bias=True, activation="linear")
        ]

model = PotNet()

def accuracy_val(model, X ,y_true):
    y_pred = np.array([model(X[i]).sigmoid().data for i in range(X.shape[0])]).round()
    correct_results = np.sum(y_pred == y_true)
    acc = correct_results/y_true.shape[0]
    acc = np.round(acc * 100)
    return acc

epochs = 100 
for epoch in range(epochs): 
    for i in range(X.shape[0]):
        output = model(X[i]).sigmoid()
        neg_output = 1-output
        target = y[i]
        bce_loss = -(target * output.log() + (1 - target) * neg_output.log())
        bce_loss.backward()
        for p in model.parameters():
            p.data += -0.001 * p.grad 
        model.zero_grad()
    acc = accuracy_val(model=model, X=X, y_true=y)
    if epoch%10 == 0:
        print(f"epoch: {epoch} |", f"loss: {bce_loss.data:.2f} |", f"accuracy: {acc}")

model.save()


from .engine import Value
import random 
from typing import Literal
from .skeleton import Module


class Neuron(Module):

    def __init__(self, nin, bias: bool = True, activation: Literal["relu", "sigmoid", "tanh", "linear"] = "linear"):
        self.w = [Value(random.uniform(-1,1)) for _ in range(nin)]
        if bias: 
            self.b = Value(random.uniform(-1,1))
        else: 
            self.b = Value(0.)
        self.activation = activation

    def __call__(self, x):
        act = sum((wi*xi for wi,xi in zip(self.w, x)), self.b)
        if self.activation == "relu":
            return act.relu()
        elif self.activation == "sigmoid":
            return act.sigmoid()
        elif self.activation == "tanh":
            return act.tanh()
        elif self.activation == "linear":
            return act 
        else:
            raise ValueError("Select between the following activation functions: relu, sigmoid, tanh, linear.")

    def parameters(self):
        return self.w + [self.b]

    def __repr__(self):
        if self.activation == "relu":
            return f"ReLu Neuron({len(self.w)})"
        elif self.activation == "sigmoid":
            return f"Sigmoid Neuron({len(self.w)})"
        elif self.activation == "tanh":
            return f"TanH Neuron({len(self.w)})"
        elif self.activation == "linear":
            return f"Linear Neuron({len(self.w)})"
        else:
            raise ValueError("Select between the following activation functions: relu, sigmoid, tanh, linear.")

    

class RNN1Neuron(Module):
 
    def __init__(self, nin, bias: bool = True, activation: Literal["relu", "sigmoid", "tanh", "linear"] = "linear"):
        self.w_xh = [Value(random.uniform(-1,1)) for _ in range(nin)]
        self.w_hh = Value(random.uniform(-1,1)) 
        if bias: 
            self.b = Value(random.uniform(-1,1))
        else: 
            self.b = Value(0.)
        self.h = Value(0)
        if activation not in ["relu", "sigmoid", "tanh", "linear"]:
            raise ValueError("Select between the following activation functions: relu, sigmoid, tanh, linear.")
        self.activation = activation
 
    def __call__(self, x):
        act = sum((wi*xi for wi,xi in zip(self.w_xh, x)), self.b) + (self.w_hh*self.h)
        if self.activation == "relu":
            self.h = act.relu()
            return act.relu()
        elif self.activation == "sigmoid":
            self.h = act.sigmoid()
            return act.sigmoid()
        elif self.activation == "tanh":
            self.h = act.tanh()
            return act.tanh()
        elif self.activation == "linear":
            self.h = act
            return act 
        else:
            pass
 
    def parameters(self):
        return self.w_xh + [self.w_hh] + [self.b]
 
    def __repr__(self):
        if self.activation == "relu":
            return f"ReLu RNN1Neuron({len(self.w)})"
        elif self.activation == "sigmoid":
            return f"Sigmoid RNN1Neuron({len(self.w)})"
        elif self.activation == "tanh":
            return f"TanH RNN1Neuron({len(self.w)})"
        elif self.activation == "linear":
            return f"Linear RNN1Neuron({len(self.w)})"
        else:
            pass



from graphviz import Digraph

def trace(root):
    nodes, edges = set(), set()
    def build(v):
        if v not in nodes:
            nodes.add(v)
            for child in v._prev:
                edges.add((child, v))
                build(child)
    build(root)
    return nodes, edges

def draw_dot(root, format='svg', rankdir='LR'):
    """
    format: png | svg | ...
    rankdir: TB (top to bottom graph) | LR (left to right)
    """
    assert rankdir in ['LR', 'TB']
    nodes, edges = trace(root)
    dot = Digraph(format=format, graph_attr={'rankdir': rankdir}) #, node_attr={'rankdir': 'TB'})
    
    for n in nodes:
        dot.node(name=str(id(n)), label = "{ data %.4f | grad %.4f }" % (n.data, n.grad), shape='record')
        if n._op:
            dot.node(name=str(id(n)) + n._op, label=n._op)
            dot.edge(str(id(n)) + n._op, str(id(n)))
    
    for n1, n2 in edges:
        dot.edge(str(id(n1)), str(id(n2)) + n2._op)
    
    return dot




from .engine import Value
from .neurons import Neuron, RNN1Neuron
from typing import Literal
from .skeleton import Module


class Layer(Module):

    def __init__(self, nin, nout, bias: bool = True, activation: Literal["relu", "sigmoid", "tanh", "linear"] = "linear"):
        self.neurons = [Neuron(nin, bias=bias, activation=activation) for _ in range(nout)]

    def __call__(self, x):
        out = [n(x) for n in self.neurons]
        return out[0] if len(out) == 1 else out

    def parameters(self):
        return [p for n in self.neurons for p in n.parameters()]

    def __repr__(self):
        return f"Layer of [{', '.join(str(n) for n in self.neurons)}]"
    

class RNN1Layer(Module):

    def __init__(self, nin, nout, activation: Literal["relu", "sigmoid", "tanh", "linear"] = "linear"):
        self.neurons = [RNN1Neuron(nin, activation=activation) for _ in range(nout)]

    def __call__(self, x):
        out = [n(x) for n in self.neurons]
        return out[0] if len(out) == 1 else out

    def parameters(self):
        return [p for n in self.neurons for p in n.parameters()]

    def __repr__(self):
        return f"RNN1 Layer of [{', '.join(str(n) for n in self.neurons)}]"


from .graph import draw_dot
from .engine import Value
from .neurons import Neuron, RNN1Neuron
from .layers import Layer, RNN1Layer
from .nn import MLP

__version__ = "0.19"

__all__ = [
    "Value",
    "draw_dot",
    "Neuron",
    "RNN1Neuron",
    "Layer",
    "RNN1Layer",
    "MLP",
]


from .engine import Value
import pickle
from typing import Literal
from .skeleton import Module


class MLP(Module):

    def __init__(self, layers):
        self.layers = layers

    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x

    def parameters(self):
        return [p for layer in self.layers for p in layer.parameters()]
    
    def __repr__(self):
        return f"MLP of [{', '.join(str(layer) for layer in self.layers)}]"
    
    def save(self, save_name: str = "weights"):
        weights = []
        for layer in self.layers:
            for neuron in layer.neurons:
                weights.append([i.data for i in neuron.w])
                weights.append(neuron.b.data)
        file_name = f'{save_name}.pkl'
        with open(file_name, 'wb') as f:
            pickle.dump(weights, f)
    
    def load(self, save_name: str = "weights"):
        file_name = f'{save_name}.pkl'
        with open(file_name, 'rb') as f:
            params = pickle.load(f)
        load_weights = []
        for param in params:
            if isinstance(param, list):
                load_param = [Value(i) for i in param]
                load_weights.append(load_param)
            else:
                load_param = Value(param)
                load_weights.append(load_param)
        i = 0
        for layer in self.layers:
            for neuron in layer.neurons:
                neuron.w = load_weights[i]
                neuron.b = load_weights[i+1]
                i += 2